/*
 * File: cmode.c
 * Date: 2011-11-21
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */


/***

This file contains the fns related to rho's CMODE mode.

***/

#include	<stdio.h>
#include	<string.h>
#include	<sys/stat.h>
#include	"rho.h"
#include	"buffer.h"
#include	"cmode.h"
#include	"cursor.h"
#include	"display.h"
#include	"file.h"
#include	"line.h"
#include	"macros.h"
#include	"messages.h"
#include	"search.h"
#include	"terminal.h"
#include	"utils.h"
#include	"window.h"




/*++

Return current column.  Stop at first non-blank given TRUE argument.

++*/
int getccol(bflg)
int bflg;

{
    int c, i, col;

    col = 0;
    for (i = 0; i < curwp->w_doto; ++i) {
        c = lgetc(curwp->w_dotp, i) & 0xFF;
        if (c != ' ' && c != '\t' && bflg)
            break;
        if (c == '\t')
            col += -(col % hardtab) + (hardtab - 1);
        else if (c < 0x20 || c == 0x7F)
            ++col;
        ++col;
    }
    return (col);
}


/*++

Insert a newline and indentation for C

++*/
int cinsert()
{
    char *cptr;                 /* string pointer into text to copy */
    int i;                      /* index into line to copy indent from */
    int llen;                   /* length of line to copy indent from */
    int bracef;                 /* was there a brace at the end of line? */
    LINE *lp;                   /* current line pointer */
    int offset;
    char ichar[NSTRING];        /* buffer to hold indent of last line */

    /* trim the whitespace before the point */
    lp = curwp->w_dotp;
    offset = curwp->w_doto;
    while (offset > 0 &&
           (lgetc(lp, offset - 1) == ' ' || lgetc(lp, offset - 1) == '\t'))
    {
        delete_backward(FALSE, 1);
        offset--;
    }

    /* check for a brace */
    bracef = (offset > 0 && lgetc(lp, offset - 1) == '{');

    /* put in the newline */
    if (insertnewline() == FALSE)
        return (FALSE);

    /* if the new line is not blank, don't indent it */
    lp = curwp->w_dotp;
    if (lused(lp) != 0)
        return (TRUE);

    /* hunt for the last non-blank line to get indentation from */
    while (lused(lp) == 0 && lp != curbp->b_linep)
        lp = lback(lp);

    /* grab a pointer to text to copy indentation from */
    cptr = ltext(lp);
    llen = lused(lp);

    /* save the indent of the last non blank line */
    i = 0;
    while ((i < llen) && (cptr[i] == ' ' || cptr[i] == '\t')
           && (i < NSTRING - 1)) {
        ichar[i] = cptr[i];
        ++i;
    }
    ichar[i] = 0;               /* terminate it */

    /* insert this saved indentation */
    insertstring(ichar);

    /* and one more tab for a brace */
    if (bracef)
        tab(FALSE, 1);

    return (TRUE);
}


/*++

Insert a brace into the text. 

++*/
int insbrace(n, c)
int n;                          /* repeat count */
int c;                          /* brace to insert (always } for now) */
{
    int ch;                     /* last character before input */
    int oc;                     /* caractere oppose a c */
    int i, count;
    int target;                 /* column brace should go after */
    LINE *oldlp;
    int oldoff;

    /* if we aren't at beginning of the line ... */
    if (curwp->w_doto != 0)

        /* scan to see if all space before this is white space */
        for (i = curwp->w_doto - 1; i >= 0; --i) {
            ch = lgetc(curwp->w_dotp, i);
            if (ch != ' ' && ch != '\t')
                return (insertnchars(n, c));
        }

    /* chercher le caractere oppose correspondant */
    switch (c) {
    case '}':
        oc = '{';
        break;
    case ']':
        oc = '[';
        break;
    case ')':
        oc = '(';
        break;
    default:
        return (FALSE);
    }

    oldlp = curwp->w_dotp;
    oldoff = curwp->w_doto;

    count = 1;
    previous_character(FALSE, 1);

    while (count > 0) {
        if (curwp->w_doto == lused(curwp->w_dotp))
            ch = '\r';
        else
            ch = lgetc(curwp->w_dotp, curwp->w_doto);

        if (ch == c)
            ++count;
        if (ch == oc)
            --count;

        previous_character(FALSE, 1);
        if (boundary(curwp->w_dotp, curwp->w_doto, REVERSE))
            break;
    }

    if (count != 0) {           /* no match */
        curwp->w_dotp = oldlp;
        curwp->w_doto = oldoff;
        return (insertnchars(n, c));
    }

    curwp->w_doto = 0;          /* debut de ligne */
    /* aller au debut de la ligne apres la tabulation */
    while ((ch = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ' || ch == '\t')
        next_character(FALSE, 1);

    /* delete back first */
    target = getccol(FALSE);    /* c'est l'indent que l'on doit avoir */
    curwp->w_dotp = oldlp;
    curwp->w_doto = oldoff;

    while (target != getccol(FALSE)) {
        if (target < getccol(FALSE))    /* on doit detruire des caracteres */
            while (getccol(FALSE) > target)
                delete_backward(FALSE, 1);
        else {                  /* on doit en inserer */
            while (target - getccol(FALSE) >= hardtab)
                insertnchars(1, '\t');
            insertnchars(target - getccol(FALSE), ' ');
        }
    }

    /* and insert the required brace(s) */
    return (insertnchars(n, c));
}



/*++

Insert a # into the text. 

++*/
int inspound()
{
    int ch;                     /* last character before input */
    int i;

    /* if we are at the beginning of the line, no go */
    if (curwp->w_doto == 0)
        return (insertnchars(1, '#'));

    /* scan to see if all space before this is white space */
    for (i = curwp->w_doto - 1; i >= 0; --i) {
        ch = lgetc(curwp->w_dotp, i);
        if (ch != ' ' && ch != '\t')
            return (insertnchars(1, '#'));
    }

    /* delete back first */
    while (getccol(FALSE) >= 1)
        delete_backward(FALSE, 1);

    /* and insert the required pound */
    return (insertnchars(1, '#'));
}



/*++

The cursor is moved to a matching fence.

++*/
int matching_fence(f, n)
int f, n;                       /* not used */
{

    LINE *oldlp;                /* original line pointer */
    int oldoff;                 /* and offset */
    int sdir;                   /* direction of search (1/-1) */
    int count;                  /* current fence level count */
    char ch;                    /* fence type to match against */
    char ofence;                /* open fence */
    char c;                     /* current character in scan */

    f = n = 0;                  /*  splint  */

    /* save the original cursor position */
    oldlp = curwp->w_dotp;
    oldoff = curwp->w_doto;

    /* get the current character */
    if (oldoff == lused(oldlp))
        ch = '\r';
    else
        ch = lgetc(oldlp, oldoff);

    /* setup proper matching fence */
    switch (ch) {
    case '(':
        ofence = ')';
        sdir = FORWARD;
        break;
    case '{':
        ofence = '}';
        sdir = FORWARD;
        break;
    case '[':
        ofence = ']';
        sdir = FORWARD;
        break;
    case ')':
        ofence = '(';
        sdir = REVERSE;
        break;
    case '}':
        ofence = '{';
        sdir = REVERSE;
        break;
    case ']':
        ofence = '[';
        sdir = REVERSE;
        break;
    default:
        TTbeep();
        return (FALSE);
    }

    /* set up for scan */
    count = 1;

    /* scan until we find it, or reach the end of file */
    while (count > 0) {
        if (sdir == FORWARD)
            next_character(FALSE, 1);
        else
            previous_character(FALSE, 1);

        if (curwp->w_doto == lused(curwp->w_dotp))
            c = '\r';
        else
            c = lgetc(curwp->w_dotp, curwp->w_doto);
        if (c == ch)
            ++count;
        if (c == ofence)
            --count;
        if (boundary(curwp->w_dotp, curwp->w_doto, sdir))
            break;
    }

    /* if count is zero, we have a match, move the sucker */
    if (count == 0) {
        curwp->w_flag |= WFMOVE;
        return (TRUE);
    }

    /* restore the current position */
    curwp->w_dotp = oldlp;
    curwp->w_doto = oldoff;
    TTbeep();
    return (FALSE);
}
