/*
 * File: replace.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */

/***

This file contains only the "replace" bits for search-and-replace,
including its message-line prompt.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"rho.h"
#include	"replace.h"
#include	"cursor.h"
#include	"display.h"
#include	"line.h"
#include	"messages.h"
#include	"search.h"
#include	"utils.h"


static int replen;              /* length of replacement string */




/*++

The prompt for query-replace.

++*/
static void mlrquery()
{
    int tcol;
    RMC *rmcptr;

    mlwrite(TEXT87);            /*"Replace '" */

    tcol = echostring(patmatch, (int) strlen(TEXT87), NPAT / 2);

    mlputs(TEXT88);             /*"' with '" */
    tcol += strlen(TEXT88);

    if (rmagical && (curwp->w_bufp->b_mode & MDMAGIC)) {
        rmcptr = &rmcpat[0];

        while (rmcptr->mc_type != MCNIL && tcol < NPAT - 8) {
            if (rmcptr->mc_type == LITSTRING)
                tcol = echostring(rmcptr->u.rstr, tcol, NPAT - 8);
            else if (rmcptr->mc_type == DITTO)
                tcol = echostring(patmatch, tcol, NPAT - 8);
            else
                tcol =
                    echostring(grpmatch[rmcptr->u.group_no], tcol,
                               NPAT - 8);
            rmcptr++;
        }
    } else
        echostring((char *) rpat, tcol, NPAT - 8);

    mlputs("'? ");
}

/*++

Delete then insert. Delete a specified length from the current point then
either insert the string directly, or make use of replacement meta-array.

++*/
static int delins(dlength, instr, use_rmc)
int dlength;
unsigned char *instr;
int use_rmc;
{
    int status;
    char *rstr;
    RMC *rmcptr;

    replen = 0;

    if ((status = deletenchars((long) dlength, FALSE)) != TRUE)
        mlwrite(TEXT93);        /*"%%ERROR while deleting" */
    else if (use_rmc && (curwp->w_bufp->b_mode & MDMAGIC)) {
        rmcptr = &rmcpat[0];
        while (rmcptr->mc_type != MCNIL && status == TRUE) {
            if (rmcptr->mc_type == LITSTRING)
                status = insertstring(rstr = rmcptr->u.rstr);
            else if (rmcptr->mc_type == DITTO)
                status = insertstring(rstr = patmatch);
            else
                status = insertstring(rstr =
                                      fixnull(grpmatch
                                              [rmcptr->u.group_no]));
            replen += strlen(rstr);
            rmcptr++;
        }
    } else {
        status = insertstring((char *) instr);
        replen = (int) strlen((char *) instr);
    }

    return (status);
}





/*++

Search for a string and replace it with another string.  Query enabled.

++*/
int query_replace(f, n)
int f;                          /* default flag */
int n;                          /* # of repetitions wanted */
{

    char c;                     /* input char for query */
    int lastoff;                /* offset (for 'u' query option) */
    int nlflag;                 /* last char of search string a <NL>? */
    int nlrepl;                 /* was a replace done on the last line? */
    int nummatch;               /* number of found matches */
    int numsub;                 /* number of substitutions */
    int oldmatchlen;            /* Closure may alter the match length. */
    int origoff;                /* and offset (for . query option) */
    int query = TRUE;           /* interactive */
    int status;                 /* success flag on pattern inputs */
    LINE *lastline;             /* position of last replace and */
    LINE *origline;             /* original "." position */

    static unsigned char *oldpatmatch = NULL;   /* allocated memory for un-do. */

    if (clexec) {
        query = FALSE;
    }


    /*
     * Don't allow this command if we are
     * in read only mode.
     */
    if (curbp->b_mode & MDVIEW)
        return (err_readonly());

    /* Check for negative repetitions.
     */
    if (f && n < 0)
        return (FALSE);

    /* Ask the user for the text of a pattern.
     */
    if ((status = readpattern(TEXT85, &pat[0], TRUE)) != TRUE)  /*     "Query replace" */
        return (status);

    /* Ask for the replacement string, and get its length.
     */
    if ((status = readpattern(TEXT86, &rpat[0], FALSE)) == ABORT)       /*"with" */
        return (status);

    /* Set up flags so we can make sure not to do a recursive
     * replace on the last line.
     */
    nlflag = (pat[matchlen - 1] == '\r');
    nlrepl = FALSE;

    /* Save original . position, reset the number of matches and
     * substitutions, and scan through the file.
     */
    origline = curwp->w_dotp;
    origoff = curwp->w_doto;
    numsub = 0;
    nummatch = 0;
    lastline = (LINE *) NULL;

    while ((f == FALSE || n > nummatch)
           && (nlflag == FALSE || nlrepl == FALSE)) {
        /* Search for the pattern.
         * If we search with a regular expression,
         * matchlen is reset to the true length of
         * the matched string.
         */
        if (magical && (curwp->w_bufp->b_mode & MDMAGIC)) {
            if (!mcscanner(FORWARD, PTBEG, 1))
                break;
        } else if (!scanner(FORWARD, PTBEG, 1))
            break;              /* all done */

        ++nummatch;             /* Increment # of matches */

        /* Check if we are on the last line.
         */
        nlrepl = (lforw(curwp->w_dotp) == curwp->w_bufp->b_linep);

        if (query) {
            /* Get the query.
             */
          pprompt:mlrquery();
          qprompt:
            /* Show the proposed place to change, and
             * update the position on the modeline if needed.
             */
            if (posflag)
                update_modeline();
            update(TRUE);
            c = tgetc();        /* and input */
            mlerase();          /* and clear it */

            /* And respond appropriately.
             */
            switch (c) {
#if	FRENCH
            case 'o':
            case 'O':
#elif GERMAN
            case 'j':
            case 'J':
#elif SPANISH
            case 's':
            case 'S':
#endif
            case 'y':          /* yes, substitute */
            case 'Y':
            case ' ':
                break;

            case 'n':          /* no, onward */
            case 'N':
                next_character(FALSE, 1);
                continue;

            case '!':          /* yes/stop asking */
                query = FALSE;
                break;

            case 'u':          /* undo last and re-prompt */
            case 'U':
                /* Restore old position.
                 */
                if (lastline == (LINE *) NULL) {
                    /* There is nothing to undo.
                     */
                    TTbeep();
                    goto pprompt;
                }
                curwp->w_dotp = lastline;
                curwp->w_doto = lastoff;
                lastline = NULL;
                lastoff = 0;

                /* Delete the new string,
                 * restore the old match.
                 */
                previous_character(FALSE, replen);
                status = delins(replen, oldpatmatch, FALSE);
                if (status != TRUE) {
                    return (status);
                }

                /* Record one less substitution,
                 * backup, save our place, and
                 * reprompt.
                 */
                --numsub;
                previous_character(FALSE, oldmatchlen);
                matchlen = oldmatchlen;
                matchline = curwp->w_dotp;
                matchoff = curwp->w_doto;
                continue;

            case '.':          /* abort! and return */
                /* restore old position */
                curwp->w_dotp = origline;
                curwp->w_doto = origoff;
                curwp->w_flag |= WFMOVE;

            case BELL:         /* abort! and stay */
                mlwrite(TEXT89);        /*"Aborted!" */
                return (FALSE);

            default:           /* bitch and beep */
                TTbeep();

            case '?':          /* help me */
                mlwrite(TEXT90);
                /*"(Y)es, (N)o, (!)Do rest, (U)ndo last, (^G)Abort, (.)Abort back, (?)Help: " */
                goto qprompt;

            }                   /* end of switch */
        }

        /* end of "if query" */
        /* if this is the point origin, flag so we a can reset it */
        if (curwp->w_dotp == origline) {
            origline = NULL;
            lastline = lback(curwp->w_dotp);
        }

        /* Delete the sucker, and insert its
         * replacement.
         */
        status = delins(matchlen, &rpat[0], rmagical);
        if (origline == NULL) {
            origline = lforw(lastline);
            origoff = 0;
        }

        if (status != TRUE) {
            return (status);
        }

        numsub++;               /* increment # of substitutions */

        /* Save our position, the match length, and the string
         * matched if we are query-replacing, as we may undo
         * the replacement. If we are not querying, check to
         * make sure that we didn't replace an empty string
         * (possible in MAGIC mode), because we'll infinite loop.
         */
        if (query) {
            lastline = curwp->w_dotp;
            lastoff = curwp->w_doto;
            oldmatchlen = matchlen;     /* Save the length for un-do. */

            if (oldpatmatch != NULL)
                free(oldpatmatch);

            oldpatmatch = malloc((size_t) (matchlen + 1));
            strcpy((char *) oldpatmatch, patmatch);

        } else if (matchlen == 0) {
            mlwrite(TEXT91);    /*"Empty string replaced, stopping." */
            return (FALSE);
        }
    }

    /* And report the results.
     */
    mlwrite(TEXT92, numsub);    /*"%d substitutions" */
    return (TRUE);
}
